@@ -17,9 +17,10 @@ You should have received a copy of the GNU General Public License
 along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
 """
 from __future__ import annotations
+
+from pathlib import Path
 from typing import *
 
-from proton import session
 from .exceptions import ProtonCryptoError, ProtonAPIError, ProtonAPIAuthenticationNeeded, ProtonAPI2FANeeded, ProtonAPIMissingScopeError, ProtonAPIHumanVerificationNeeded
 from .srp import User as PmsrpUser
 from .environments import Environment
@@ -29,6 +30,8 @@ import asyncio
 import base64
 import random
 
+from ..utils import ExecutionEnvironment
+
 SRP_MODULUS_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----
 
 xjMEXAHLgxYJKwYBBAHaRw8BAQdAFurWXXwjTemqjD7CXjXVyKf0of7n9Ctm
@@ -117,7 +120,8 @@ class Session:
 
     async def async_api_request(self, endpoint,
         jsondata=None, data=None, additional_headers=None,
-        method=None, params=None, no_condition_check=False):
+        method=None, params=None, no_condition_check=False,
+        return_raw=False):
         """Do an API request.
 
         This call can return any of the exceptions defined in :mod:`proton.session.exceptions`.
@@ -141,9 +145,13 @@ class Session:
         :type params: str, dict or iterable
         :param no_condition_check: Internal flag to disable locking, defaults to False
         :type no_condition_check: bool, optional
+        :param return_raw: If set to true, the function will return a :class:`RawResponse` object
+            instead of a dict, this class contains the header and status information along with the
+            json response.
+        :type return_raw: bool, optional
 
-        :return: Deserialized JSON reply
-        :rtype: dict
+        :return: Deserialized JSON reply or a RawResponse object if return_raw is True
+        :rtype: dict | RawResponse
         """
 
         # We might need to loop
@@ -153,7 +161,8 @@ class Session:
             attempts -= 1
             try:
                 refresh_revision_at_start = self.__refresh_revision
-                return await self.__async_api_request_internal(endpoint, jsondata, data, additional_headers, method, params, no_condition_check)
+                return await self.__async_api_request_internal(endpoint, jsondata, data, additional_headers, method, params, no_condition_check,
+                                                               return_raw=return_raw)
             except ProtonAPIError as e:
                 stored_exception = e
                 # We have a missing scope.
@@ -426,7 +435,57 @@ class Session:
 
     async def async_human_verif_provide_token(self, method, token):
         pass
-    
+
+
+    async def async_fork(self, child_client_id: str, payload: str=None, independent: int=0, user_code: int=None, no_condition_check=False) -> str:
+        """Try to fork the current session with parameters and return the fork selector if successful.
+
+        :param child_client_id: The clientID that the client creating the child session will use.
+        :type child_client_id: str
+        :param payload: Will be downloaded by the child client, and can contain encrypted sensitive information.
+        :type payload: str, optional
+        :param independent: set to 1 if the newly forked session has to be independent, else 0.
+        :type independent: int, optional
+        :param user_code: If provided, the selector will not be randomly chosen, but rather the one already returned will be used.
+        :type user_code: int, optional
+        :param no_condition_check: Internal flag to disable locking, defaults to False
+        :type no_condition_check: bool, optional
+
+        :return selector: Fork selector to be used by method `async_import_fork()`
+        :rtype: str
+        """
+        data = {'ChildClientID' : child_client_id, 'Independent': independent}
+        if payload is not None:
+            data['Payload'] = payload
+        if user_code is not None:
+            data['UserCode'] = user_code
+        ret = await self.async_api_request('/auth/v4/sessions/forks', data, method='POST', no_condition_check=no_condition_check)
+        return ret['Selector']
+
+
+    async def async_import_fork(self, selector: str, no_condition_check=False) -> str:
+        """Try to import the session fork specified by the selector.
+
+        :param selector: Obtained via a successful call to `async_fork()`
+        :type selector: str
+        :param no_condition_check: Internal flag to disable locking, defaults to False
+        :type no_condition_check: bool, optional
+
+        :return payload: The payload sent by the parent session when initiating the fork.
+        :rtype: str
+        """
+        self._requests_lock(no_condition_check)
+        try:
+            ret = await self.async_api_request(f"/auth/v4/sessions/forks/{selector}", method='GET', no_condition_check=True)
+            self.__UID = ret['UID']
+            self.__RefreshToken = ret['RefreshToken']
+            self.__AccessToken = ret['AccessToken']
+            self.__Scopes = ret['Scopes']
+            return ret['Payload']
+        finally:
+            self._requests_unlock(no_condition_check)
+        
+
 
     # Wrappers to provide non-asyncio API
     api_request = sync_wrapper(async_api_request)
@@ -437,6 +496,8 @@ class Session:
     lock = sync_wrapper(async_lock)
     human_verif_request_code = sync_wrapper(async_human_verif_request_code)
     human_verif_provide_token = sync_wrapper(async_human_verif_provide_token)
+    fork = sync_wrapper(async_fork)
+    import_fork = sync_wrapper(async_import_fork)
 
     def register_persistence_observer(self, observer: object):
         """Register an observer that will be notified of any persistent state change of the session
@@ -702,7 +763,8 @@ class Session:
     async def __async_api_request_internal(
         self, endpoint,
         jsondata=None, data=None, additional_headers=None,
-        method=None, params=None, no_condition_check=False
+        method=None, params=None, no_condition_check=False,
+        return_raw=False
     ):
         """Internal function to do an API request (without clever exception handling and retrying). 
         See :meth:`async_api_request` for the parameters specification."""
@@ -713,13 +775,20 @@ class Session:
             raise RuntimeError("Could not instanciate a transport, are required dependencies installed?")
 
         await self._requests_wait(no_condition_check)
-        return await self.__transport.async_api_request(endpoint, jsondata, data, additional_headers, method, params)
+        return await self.__transport.async_api_request(endpoint, jsondata, data, additional_headers, method, params,
+                                                        return_raw=return_raw)
 
     def _verify_modulus(self, armored_modulus) -> bytes:
         if self.__gnupg_for_modulus is None:
             import gnupg
-            # Verify modulus
-            self.__gnupg_for_modulus = gnupg.GPG()
+            # The modulus key is imported in a separate GPG keyring (rather than on the user's
+            # default one) by modifying the gnupg home directory. The reason for doing this is
+            # that we received crash reports due to users messing up with their default GPG
+            # keyring permissions. In this case, python-gnupg fails silently when importing
+            # Proton's modulus key, which then causes the signature verification errors.
+            proton_gnupg_dir = Path(ExecutionEnvironment().path_cache) / "gnupg"
+            proton_gnupg_dir.mkdir(exist_ok=True)
+            self.__gnupg_for_modulus = gnupg.GPG(gnupghome=proton_gnupg_dir)
             self.__gnupg_for_modulus.import_keys(SRP_MODULUS_KEY)
 
         # gpg.decrypt verifies the signature too, and returns the parsed data.
